"use strict";
this.name        = "NPC_Energy_Units";
this.author      = "Dybal";
this.copyright   = "2020 Dybal";
this.licence     = "CC-BY-NC-SA 4.0";
this.description = "NPC Energy Units script";
this.version     = "1.0";

this._energyUnits = [ "EQ_ENERGY_UNIT", "EQ_NAVAL_ENERGY_UNIT" ];
this._rechargeMultipliers = {}; // recharge multipliers for each energy unit, initialized at startUp
this._spawnedShipsTimers = []; // will hold timer references for the spawned ships

this._debug = true;
this._logging = true;

//
// Event Handlers
//

//---------------------------------------------------------------------------------//
this.startUp = function _npc_energy_units_startUp() {
    var eu = this._energyUnits;
    var rechargeMultipliers = this._rechargeMultipliers;
    var eqInfo, eqKey, i = eu.length;
    while (i--) {
        eqKey = eu[i];
        eqInfo = EquipmentInfo.infoForKey(eqKey);
        if (eqInfo.scriptInfo && eqInfo.scriptInfo.oolite_energy_recharge_multiplier)
            rechargeMultipliers[eqKey] = parseFloat(eqInfo.scriptInfo.oolite_energy_recharge_multiplier);
    }
    if (this._debug) log(this.name, "Recharge multipliers: "+JSON.stringify(this._rechargeMultipliers));

    if (worldScripts.NPC_Equipment_Damage)
        worldScripts.NPC_Equipment_Damage.NPC_equipmentDamagedHandlers.push(this.$energyUnitDamaged);
}

//----------------------------------------------------------------------------------------------//
this.shipSpawned = function _npc_energy_units_shipSpawned(ship) {
    if (ship.isBoulder || ship.isCargo || ship.isRock || !ship.isPiloted) return;
    var _spawnedShipsTimers = this._spawnedShipsTimers;
    var id = _spawnedShipsTimers.length; // array indices start at 0
    this._spawnedShipsTimers[id] = new Timer(this, this.$setSpawnedShip.bind(this, id, ship), 0, 0);
}


//
//  Functions for OXPs to use
//

//-----------------------------------------------------------------------------------//
this.$getShipEnergyUnit = function _npc_energy_units_getShipEnergyUnit(ship) {
    var that = _npc_energy_units_getShipEnergyUnit;
    var ws = (that.ws = that.ws || worldScripts.NPC_Energy_Units);
    var eus = ws._energyUnits;
    var i = eus.length;
    var eqKey;

    if (!ship|| !ship.isValid) {
        log(ws.name, "$getShipEnergyUnit ERROR: invalid ship");
        return null;
    }
    while (i--) {
        eqKey = eus[i];
        if (["EQUIPMENT_OK", "EQUIPMENT_DAMAGED"].indexOf(ship.equipmentStatus(eqKey)) !== -1)
            return eqKey;
    }
    return null;
}

//-----------------------------------------------------------------------------------//
this.$setupEnergyUnits = function _npc_energy_units_setupEnergyUnits(ship) {
    var that = _npc_energy_units_setupEnergyUnits;
    var ws = (that.ws = that.ws || worldScripts.NPC_Energy_Units);
    var eus = ws._energyUnits;
    var rechargeMultipliers = ws._rechargeMultipliers;
    var eqKey, i = eus.length;
    var shipdata = Ship.shipDataForKey(ship.dataKey);

    if (!ship|| !ship.isValid) {
        log(ws.name, "$setupEnergyUnits ERROR: invalid ship");
        return false;
    }

    if (!shipdata.energy_recharge_rate) {
        log(ws.name, ship.displayName+": $setupEnergyUnits ERROR setting up energy units: factory recharge rate not found in shipdata for key "+ship.dataKey);
        return false;
    }
    var original_recharge_rate = parseFloat(shipdata.energy_recharge_rate);

    while (i--) {
        eqKey = eus[i];
        if (ship.equipmentStatus(eqKey) === "EQUIPMENT_OK" && rechargeMultipliers[eqKey]) {
            let eqInfo = EquipmentInfo.infoForKey(eqKey);
            let old_recharge = ship.energyRechargeRate;
            // we can't use the oolite-equipment-control strategy because it isn't idempotent and we need it to be...
            // this function might be called several times on the same ship by different OXPs
            let eboost = rechargeMultipliers[eqKey] * original_recharge_rate;
            ship.energyRechargeRate = eboost;
            if (ws._logging) log(ws.name, ship.displayName+": "+eqInfo.name+" enabled, energy recharge "+old_recharge.toFixed(1)+"->"+ship.energyRechargeRate.toFixed(1));
            return true;
        }
    }
    return false;
}

//-----------------------------------------------------------------------------------//
this.$energyUnitDamaged = function _npc_energy_units_energyUnitDamaged(eqKey) { // MUST be called with the ship.script as 'this'
    var that = _npc_energy_units_energyUnitDamaged;
    var ws = (that.ws = that.ws || worldScripts.NPC_Energy_Units);
    var rechargeMultipliers = ws._rechargeMultipliers;
    var eqInfo = EquipmentInfo.infoForKey(eqKey);
    var shipdata = Ship.shipDataForKey(ship.dataKey);
    var ship = this.ship;

    if (!ship || !ship.isValid) {
        log(ws.name, "$energyUnitDamaged ERROR: invalid ship:"+ship);
        return false;
    }
    if (!rechargeMultipliers[eqKey]) {
        log(ws.name, ship.displayName+": $energyUnitDamaged ERROR: invalid energy unit key: "+eqKey);
        return false;
    }
    if (!shipdata.energy_recharge_rate) {
        log(ws.name, ship.displayName+": $energyUnitDamaged ERROR disabling energy unit: factory recharge rate not found in shipdata for key "+ship.dataKey);
        return false;
    }
    var original_recharge_rate = parseFloat(shipdata.energy_recharge_rate);
    
    if (ship.equipmentStatus(eqKey) !== "EQUIPMENT_OK") {
        let old_recharge = ship.energyRechargeRate;
        // we can't use the oolite-equipment-control strategy because it isn't idempotent and we need it to be...
        // this function might be called several times on the same ship by different OXPs
        ship.energyRechargeRate = original_recharge_rate;
        if (ws._logging) log(ws.name, ship.displayName+": "+eqInfo.name+" disabled, energy recharge "+old_recharge.toFixed(1)+"->"+ship.energyRechargeRate.toFixed(1));
        return true;
    }
    return false;
}


//
// Internal Functions
//

//----------------------------------------------------------------------------------------------//
this.$setSpawnedShip = function _npc_energy_units_setSpawnedShip(id, ship) { 
    var that = _npc_energy_units_setSpawnedShip;
    var _spawnedShipsTimers = (that._spawnedShipsTimers =  that._spawnedShipsTimers || this._spawnedShipsTimers);

    // energy units awarding
    var eqKey = this.$getShipEnergyUnit(ship);
    if (!eqKey) {
        var rnd = Math.random();
        if (this._debug) log(this.name, ship.displayName+": accuracy: "+ship.accuracy.toFixed(1)+", dice roll:"+rnd.toFixed(2));

        if (rnd >= (15 - ship.accuracy) / 25 + (ship.energyRechargeRate / 7)) {
            if (ship.awardEquipment("EQ_NAVAL_ENERGY_UNIT")) eqKey = "EQ_NAVAL_ENERGY_UNIT";
            if (eqKey && this._logging) log(this.name, "Awarding Naval Energy Unit to "+ship.displayName);
        } else if (rnd >= (15 - ship.accuracy) / 35 + (ship.energyRechargeRate / 10)) {
            if (ship.awardEquipment("EQ_ENERGY_UNIT")) eqKey = "EQ_ENERGY_UNIT";
            if (eqKey && this._logging) log(this.name, "Awarding Extra Energy Unit to "+ship.displayName);
        }
    }
    // energy units setup
    if (eqKey) this.$setupEnergyUnits(ship);

    // housekeeping
    delete _spawnedShipsTimers[id];
    _spawnedShipsTimers[id] = null;
    var i = _spawnedShipsTimers.length;
    var count = 0;
    while (i--) if (_spawnedShipsTimers[i]) count++;
    if (count === 0) {
        // no active timers, clean the array
        if (this._debug) log(this.name, "No more NPC setup timers running, cleaning the timers array");
        _spawnedShipsTimers.length = 0;
    } else if (this._debug) log(this.name, count+" NPC setup timers still running");
}
